/*
 * Workspaces.cpp
 * Open BeOS version alpha 1 by Francois Revol revol@free.fr
 *
 * Workspaces window trick found by Michael "Minox" Paine.
 * (using B_ALL_WORKSPACES as flags in BWindow)
 * Found out that using 0xffffffff as Flags was causing the window not to close on Alt-W
 * hey Workspaces get Flags of Window 0
 * gives 0x00008080 which makes it.
 */

#include <Application.h>
#include <Alert.h>
#include <File.h>
#include <PopUpMenu.h>
#include <MenuItem.h>
#include <Window.h>
#include <Screen.h>
#include <Entry.h>
#include <Roster.h>
#include <Path.h>
#include <FindDirectory.h>
#include <MessageRunner.h>

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>

#define WS_FLAG_NO_TAB		0x00000001
#define WS_FLAG_NO_BORDER	0x00000002
#define WS_FLAG_TOPMOST		0x00000004
#define WS_FLAG_ACCEPT_FIRST_CLICK	0x00000008
/* UNIMPLEMENTED */
#define WS_FLAG_GROW	0x00000010

// 1 pixel border
//#define BORDERLESS_STYLE B_BORDERED_WINDOW_LOOK
// no border at all
#define BORDERLESS_STYLE B_NO_BORDER_WINDOW_LOOK

//#define NOTAB_STYLE B_FLOATING_WINDOW_LOOK
#define NOTAB_STYLE B_MODAL_WINDOW_LOOK

// if defined:
// poll for right mouse button pressed then released to popup,
// if not defined:
// only popup the menu on ESC or SPACE
#define USE_POLLING_FOR_POPUP

class WorkspacesPreferences {
	public:
		WorkspacesPreferences();
		virtual ~WorkspacesPreferences();

		BRect WindowFrame() const { return fWindowFrame; }
		BRect ScreenFrame() const { return fScreenFrame; }
		BRect PreviousWindowFrame() const { return previousWindowFrame; }
		BRect PreviousScreenFrame() const { return previousScreenFrame; }
		bool LoadedDifferentResolution() const { return loadedDifferentResolution; }
		uint32 Flags() const { return fFlags; };
		uint32 Corner() const { return fCorner; };
	
		void UpdateScreenFrame();
		void SetWindowFrame(BRect);
		void SetPreviousScreenFrame(BRect);
		void SetPreviousWindowFrame(BRect);
		void SetLoadedDifferentResolution(bool);
		void SetCorner(uint32 corner);
		void SetFlags(uint32 flags);

	private:
		static const char kWorkspacesSettingFile[];
		static const char kWorkspacesResolutionFile[];
		BRect fWindowFrame, fScreenFrame;
		BRect previousWindowFrame, previousScreenFrame;
		bool loadedDifferentResolution;
		uint32 fFlags;
		uint32 fCorner;
};

class WorkspacesWindow : public BWindow {
	public:
		WorkspacesWindow(WorkspacesPreferences *fPreferences);
		virtual ~WorkspacesWindow();
	
		virtual void ScreenChanged(BRect frame,color_space mode);
		virtual void FrameMoved(BPoint origin);
		virtual void FrameResized(float width, float height);
		virtual void Zoom(BPoint origin, float width, float height);
		virtual void MessageReceived(BMessage *msg);
		virtual bool QuitRequested(void);
		virtual void WorkspaceActivated(int32 ws, bool state);
		
		void Popup(BPoint hotspot);

	private:
		WorkspacesPreferences *fPreferences;
/**/		BView *utilView;
		uint32 lastButtons;
/**/
};

class WorkspacesApp : public BApplication {
	public:
		WorkspacesApp();
		virtual ~WorkspacesApp();

		virtual void AboutRequested(void);
		virtual void ArgvReceived(int32 argc, char **argv);
		virtual void ReadyToRun(void);
		virtual bool QuitRequested(void);
		virtual void Usage(void);
		virtual void Version(void);
		virtual void SuggestHelpMessage(void);

	private:
		static const char kWorkspacesAppSig[];
		WorkspacesWindow *fWorkspacesWindow;
		char *prog_name;
};

const char WorkspacesApp::kWorkspacesAppSig[] = "application/x-vnd.Be-WORK";
const char WorkspacesPreferences::kWorkspacesSettingFile[] = "Workspace_data";
const char WorkspacesPreferences::kWorkspacesResolutionFile[] = "Workspace_resolution";

WorkspacesPreferences::WorkspacesPreferences()
{
	UpdateScreenFrame();

	BPath workspaceDataPath;
	BScreen screen;
	fFlags = 0;
	fCorner = 0;

	if (find_directory(B_USER_SETTINGS_DIRECTORY,&workspaceDataPath) == B_OK) {
		workspaceDataPath.Append(kWorkspacesSettingFile);
		BFile workspaceDataFile(workspaceDataPath.Path(), B_READ_ONLY);
		if (workspaceDataFile.InitCheck() == B_OK 
			&& workspaceDataFile.Read(&fScreenFrame, sizeof(BRect)) == sizeof(BRect)) {
			if ( fScreenFrame != screen.Frame() )
				loadedDifferentResolution = true;
			else 
				loadedDifferentResolution = false;
			if (workspaceDataFile.Read(&fWindowFrame, sizeof(BRect)) == sizeof(BRect)){
				if (workspaceDataFile.Read(&fFlags, sizeof(uint32)) == sizeof(uint32))
					if (workspaceDataFile.Read(&fCorner, sizeof(uint32)) == sizeof(uint32))
						return;
				return;
			}	
		}
	}
	fWindowFrame = fScreenFrame;
	fWindowFrame.OffsetBy(-10, -10);
	fWindowFrame.left = fWindowFrame.right - 160;
	fWindowFrame.top = fWindowFrame.bottom - 120;

}


WorkspacesPreferences::~WorkspacesPreferences()
{	
	// write Workspace_data
	BPath workspaceDataPath;
	if (find_directory(B_USER_SETTINGS_DIRECTORY,&workspaceDataPath) < B_OK)
		return;

	workspaceDataPath.Append(kWorkspacesSettingFile);

	BFile workspaceDataFile(workspaceDataPath.Path(),
		 B_WRITE_ONLY | B_CREATE_FILE);
	if (workspaceDataFile.InitCheck() == B_OK)
		workspaceDataFile.Write(&fScreenFrame, sizeof(BRect));
		workspaceDataFile.Write(&fWindowFrame, sizeof(BRect));
		workspaceDataFile.Write(&fFlags, sizeof(uint32));
		workspaceDataFile.Write(&fCorner, sizeof(uint32));
		

}

void WorkspacesPreferences::SetLoadedDifferentResolution(bool b) 
{
	loadedDifferentResolution = b;
}

void
WorkspacesPreferences::SetPreviousScreenFrame(BRect f)
{
	previousScreenFrame = f;
}

void
WorkspacesPreferences::SetPreviousWindowFrame(BRect f)
{
	previousWindowFrame = f;
}

void
WorkspacesPreferences::UpdateScreenFrame()
{
	BScreen screen;
	fScreenFrame = screen.Frame();
}


void
WorkspacesPreferences::SetWindowFrame(BRect f)
{
	fWindowFrame = f;

	UpdateScreenFrame();
}

void
WorkspacesPreferences::SetFlags(uint32 flags)
{
//printf("flags: %08lx -> %08lx\n", fFlags, flags);
	fFlags = flags;
}

void
WorkspacesPreferences::SetCorner(uint32 corner)
{
//printf("corner: %08lx -> %08lx\n", fCorner, corner);
	fCorner = corner;
}


//	#pragma mark -

// here is the trick :)
#define B_WORKSPACES_WINDOW 0x00008000

WorkspacesWindow::WorkspacesWindow(WorkspacesPreferences *preferences)
	: BWindow(preferences->WindowFrame(), "Workspaces", B_TITLED_WINDOW_LOOK,
 			B_NORMAL_WINDOW_FEEL, B_WORKSPACES_WINDOW | B_AVOID_FRONT, B_ALL_WORKSPACES)
{
	fPreferences = preferences;
	// CLI arguments, if used, will override these Look ,Feel, and Flags.
	if (fPreferences->Flags() & WS_FLAG_NO_BORDER)
		SetLook(BORDERLESS_STYLE);
	else if (fPreferences->Flags() & WS_FLAG_NO_TAB)
		SetLook(NOTAB_STYLE);
	else 
		SetLook(B_TITLED_WINDOW_LOOK);

	if (fPreferences->Flags() & WS_FLAG_TOPMOST)
		SetFeel((window_feel)(Feel() ^ B_FLOATING_ALL_WINDOW_FEEL));
	if (fPreferences->Flags() & WS_FLAG_ACCEPT_FIRST_CLICK)
		SetFlags(Flags() ^ B_WILL_ACCEPT_FIRST_CLICK);
/**/
	lastButtons = 0;
	utilView = new BView(BRect(0, 0, -1, -1), "util view", B_FOLLOW_NONE, 0);
	AddChild(utilView);
/**/
#ifdef USE_POLLING_FOR_POPUP
	BMessenger msgr(this);
	new BMessageRunner(msgr, new BMessage('PulS'), 500000);
#endif /* USE_POLLING_FOR_POPUP */

	// record initial resolution and window position as old,
	// for first call to ScreenChanged()
	fPreferences->SetPreviousScreenFrame(fPreferences->ScreenFrame()); 
	fPreferences->SetPreviousWindowFrame(Frame());  
	if ( fPreferences->LoadedDifferentResolution() ) {  
		BScreen screen;
		fPreferences->SetPreviousWindowFrame(fPreferences->WindowFrame());
		ScreenChanged(screen.Frame(), screen.ColorSpace());
	}	
}


WorkspacesWindow::~WorkspacesWindow()
{
	delete fPreferences;
}


void
WorkspacesWindow::ScreenChanged(BRect rect, color_space mode)
{
	//`rect`  		new screen dimensions
	//`Frame()`		new ws window location

	// need to be stored locally, since PreviousWindowFrame() will get modified.
	// well, it would if FrameMoved() weren't called before ScreenChanged().
	BRect oldScreenFrame = fPreferences->PreviousScreenFrame();
	BRect oldWindowFrame = fPreferences->PreviousWindowFrame();

	// adjust horizontal position
	if (oldWindowFrame.right > oldScreenFrame.right/2)  
		MoveTo(rect.right - (oldScreenFrame.right - oldWindowFrame.left),
			oldWindowFrame.top); 

	// adjust vertical position
	if (oldWindowFrame.bottom > oldScreenFrame.bottom/2) 
		MoveTo(Frame().left, 
			rect.bottom - (oldScreenFrame.bottom - oldWindowFrame.top));

	// call these at the end of the function, so it's ready for next call.
	// FLAW: moving the workspace window after a ScreenChanged will use these values.
	fPreferences->SetPreviousScreenFrame(rect);  
	fPreferences->SetPreviousWindowFrame(Frame());
}


void
WorkspacesWindow::FrameMoved(BPoint origin)
{
	fPreferences->SetWindowFrame(Frame());
	// When switching workspaces, FrameMoved is called before ScreenChanged.
	//  prevents ScreenChange correctly repositioning after a 
	//  workspaces window repositition.
	// Must find a way to SetPreviousWindowFrame(Frame()) in FrameMoved(),
	//  but not when FrameMoved is called upon switching workspaces.
	//fPreferences->SetPreviousWindowFrame(Frame());
}


void
WorkspacesWindow::FrameResized(float width, float height)
{
	fPreferences->SetWindowFrame(Frame());
	fPreferences->SetPreviousWindowFrame(Frame());	
}


void 
WorkspacesWindow::Zoom(BPoint origin, float width, float height)
{
	BScreen screen;
	origin = screen.Frame().RightBottom();
	origin.x -= 10 + fPreferences->WindowFrame().Width();
	origin.y -= 10 + fPreferences->WindowFrame().Height();

	MoveTo(origin);
}


void
WorkspacesWindow::MessageReceived(BMessage *msg)
{
	int32 key;
	entry_ref ref;
	uint32 corner;
	
	switch (msg->what) {
	case 'DATA': // Drop from Tracker
		for (int i = 0; (msg->FindRef("refs", i, &ref) == B_OK); i++)
		{
//		BMessage m(B_OPEN_IN_WORKSPACE);
//		m.AddInt32("be:workspace", 3);
//		be_roster->Launch(&ref, &m);
			be_roster->Launch(&ref);
		}
		break;
	case B_KEY_UP:
		if (msg->FindInt32("raw_char", &key) == B_OK && (key == ' ' || key == 033)) { // space key
			BPoint hotspot(Frame().LeftTop());
			Popup(hotspot);
		} else
			BWindow::MessageReceived(msg);
		break;
	case 'wsBn':
		fPreferences->SetFlags(fPreferences->Flags() & ~(WS_FLAG_NO_TAB | WS_FLAG_NO_BORDER));
		SetLook(B_TITLED_WINDOW_LOOK);
		if (corner) // force update
			WorkspaceActivated(current_workspace(), true);
		break;
	case 'wsBt':
		fPreferences->SetFlags(fPreferences->Flags() ^ WS_FLAG_NO_TAB);
		if (fPreferences->Flags() & WS_FLAG_NO_BORDER)
			SetLook(BORDERLESS_STYLE);
		else if (fPreferences->Flags() & WS_FLAG_NO_TAB)
			SetLook(NOTAB_STYLE);
		else 
			SetLook(B_TITLED_WINDOW_LOOK);
		if (corner) // force update
			WorkspaceActivated(current_workspace(), true);
		break;
	case 'wsBl':
		fPreferences->SetFlags(fPreferences->Flags() ^ WS_FLAG_NO_BORDER);
		if (fPreferences->Flags() & WS_FLAG_NO_BORDER)
			SetLook(BORDERLESS_STYLE);
		else if (fPreferences->Flags() & WS_FLAG_NO_TAB)
			SetLook(NOTAB_STYLE);
		else 
			SetLook(B_TITLED_WINDOW_LOOK);
		if (corner) // force update
			WorkspaceActivated(current_workspace(), true);
		break;
	case 'wsTM':
		SetFeel((window_feel)(Feel() ^ B_FLOATING_ALL_WINDOW_FEEL));
		fPreferences->SetFlags(fPreferences->Flags() ^ WS_FLAG_TOPMOST);
		break;
	case 'wsAc':
		SetFlags(Flags() ^ B_WILL_ACCEPT_FIRST_CLICK);
		fPreferences->SetFlags(fPreferences->Flags() ^ WS_FLAG_ACCEPT_FIRST_CLICK);
		break;
	case 'wsC0':
	case 'wsC1':
	case 'wsC2':
	case 'wsC3':
	case 'wsC4':
		corner = msg->what - 'wsC0';
		fPreferences->SetCorner(corner);
		if (corner) // force update
			WorkspaceActivated(current_workspace(), true);
		break;
#ifdef USE_POLLING_FOR_POPUP
	case 'PulS': /* app_server swallows mouse events, so we poll them */
	{
		if (!IsActive())
			return;
		uint32 buttons;
		BPoint hotspot;
		utilView->GetMouse(&hotspot, &buttons);
		if (!(buttons & ~B_PRIMARY_MOUSE_BUTTON) && (lastButtons & ~B_PRIMARY_MOUSE_BUTTON)) {
			utilView->ConvertToScreen(&hotspot);
			Popup(hotspot);
		}
		lastButtons = buttons;
	}
#endif /* USE_POLLING_FOR_POPUP */
	default:
		BWindow::MessageReceived(msg);
	}
		
}


bool
WorkspacesWindow::QuitRequested(void)
{
	be_app->PostMessage(B_QUIT_REQUESTED);
	return true;
}

void
WorkspacesWindow::WorkspaceActivated(int32 ws, bool state)
{
	uint32 corner = fPreferences->Corner();
	if (!corner)
		return;
	float xDelta = 0;
	float yDelta = 0;
	int borderWidth = (fPreferences->Flags() & WS_FLAG_NO_BORDER)?0:4;
	BScreen bs;
	BRect sFrame = bs.Frame();
	if (corner < 3) // top
		yDelta = sFrame.top - (Frame().top - borderWidth);
	else // bottom
		yDelta = sFrame.bottom - (Frame().bottom + borderWidth);
	if (corner % 2) // left
		xDelta = sFrame.left - (Frame().left - borderWidth);
	else // right
		xDelta = sFrame.right - (Frame().right + borderWidth);
	if (xDelta || yDelta)
		MoveBy(xDelta, yDelta);
}

void
WorkspacesWindow::Popup(BPoint hotspot)
{
	BPopUpMenu *m;
	BMenu *tmpm;
	BMenuItem *tmpi;
	
	m = new BPopUpMenu("Workspaces popup menu", false, false);
	m->SetFont(be_plain_font);

	tmpm = new BMenu("Window Border");
	tmpm->SetFont(be_plain_font);
	
	tmpi = new BMenuItem("Normal", new BMessage('wsBn'));
	tmpi->SetMarked((fPreferences->Flags() & (WS_FLAG_NO_TAB | WS_FLAG_NO_BORDER)) == 0);
	tmpi->SetTarget(NULL, this);
	tmpm->AddItem(tmpi);

	tmpi = new BMenuItem("No Tab", new BMessage('wsBt'));
	tmpi->SetMarked(fPreferences->Flags() & WS_FLAG_NO_TAB);
	tmpi->SetTarget(NULL, this);
	tmpm->AddItem(tmpi);

	tmpi = new BMenuItem("None", new BMessage('wsBl'));
	tmpi->SetMarked(fPreferences->Flags() & WS_FLAG_NO_BORDER);
	tmpi->SetTarget(NULL, this);
	tmpm->AddItem(tmpi);

	tmpi = new BMenuItem(tmpm);
	m->AddItem(tmpi);

	tmpi = new BMenuItem("Stay On Top", new BMessage('wsTM'));
	tmpi->SetMarked((Feel() & B_FLOATING_ALL_WINDOW_FEEL) != 0);
	tmpi->SetTarget(NULL, this);
	m->AddItem(tmpi);
	
	tmpi = new BMenuItem("Accept first click", new BMessage('wsAc'));
	tmpi->SetMarked(Flags() & B_WILL_ACCEPT_FIRST_CLICK);
	tmpi->SetTarget(NULL, this);
	m->AddItem(tmpi);
				
	tmpm = new BMenu("Stick To Corner");
	tmpm->SetFont(be_plain_font);
	
	tmpi = new BMenuItem("None", new BMessage('wsC0'));
	tmpi->SetMarked(fPreferences->Corner() == 0);
	tmpi->SetTarget(NULL, this);
	tmpm->AddItem(tmpi);
	tmpi = new BMenuItem("Top Left", new BMessage('wsC1'));
	tmpi->SetMarked(fPreferences->Corner() == 1);
	tmpi->SetTarget(NULL, this);
	tmpm->AddItem(tmpi);
	tmpi = new BMenuItem("Top Right", new BMessage('wsC2'));
	tmpi->SetMarked(fPreferences->Corner() == 2);
	tmpi->SetTarget(NULL, this);
	tmpm->AddItem(tmpi);
	tmpi = new BMenuItem("Bottom Left", new BMessage('wsC3'));
	tmpi->SetMarked(fPreferences->Corner() == 3);
	tmpi->SetTarget(NULL, this);
	tmpm->AddItem(tmpi);
	tmpi = new BMenuItem("Bottom Right", new BMessage('wsC4'));
	tmpi->SetMarked(fPreferences->Corner() == 4);
	tmpi->SetTarget(NULL, this);
	tmpm->AddItem(tmpi);
	
	tmpi = new BMenuItem(tmpm);
	m->AddItem(tmpi);

	m->Go(hotspot, true, true, BRect(0,0,-1,-1)/*ConvertToScreen(Bounds())*/, true);
}


//	#pragma mark -


WorkspacesApp::WorkspacesApp() : BApplication(kWorkspacesAppSig)
{
	fWorkspacesWindow = new WorkspacesWindow(new WorkspacesPreferences());
}


WorkspacesApp::~WorkspacesApp()
{
}


void
WorkspacesApp::AboutRequested(void)
{
	// blocking !! 
	// the original does the same by the way =)
	(new BAlert("about", "OpenBeOS Workspaces\n\tby François Revol, Axel Dörfler.\n\noriginal Be version by Robert Polic", "Big Deal"))->Go();
}

void
WorkspacesApp::Usage(void)
{
	(void)fprintf(stdout, 
"%s OBOS (http://www.openbeos.org/)

Usage: %s [OPTION] [n] [APP]

  --notitle   	  		title bar removed.  border and resize kept
  --noborder   	  		title, border, and resize removed
  --acceptfirstclick  	forces workspace switch on first click
  --avoidfocus   		prevents the window from being the target of keyboard events
  --alwaysontop			keeps window on top
  --help      			display this help and exit
  --version   			output version information and exit

  [n] 		switch to Workspace[0-31]
  [APP] 	start APP with Workspaces
  			if n is specifed, APP loads on Workspace[0-31]
  			if not specified, APP loads on Workspace[0]
  			
  Note:
  To access Workspaces context menu,  press ESC or SPACE while Workspaces has focus.
  Options set via the context menu will persist from session to session.
  The context menu is inaccessible when --avoidfocus is used.
  Command line arguments will override the context menu's options. 
", prog_name, prog_name	);
	Quit();
}

void
WorkspacesApp::Version(void)
{
	(void)fprintf(stdout, "%s OBOS (http://www.openbeos.org/)\n", prog_name);
	Quit();
}

void
WorkspacesApp::SuggestHelpMessage(void)
{
	(void)fprintf(stdout, "Try `%s --help' for more information.\n", prog_name);
	Quit();
}
void
WorkspacesApp::ArgvReceived(int32 argc, char **argv)
{
	prog_name = argv[0];
	for (int i=1;  i < argc;  i++)
	{
		if (argv[i][0] == '-' && argv[i][1] == '-') {
			// evaluate --arguments
			if (! strcmp(argv[i], "--notitle")) {
				fWorkspacesWindow->SetLook(B_MODAL_WINDOW_LOOK);
				fWorkspacesWindow->SetFlags(B_OUTLINE_RESIZE);
			} else if (! strcmp(argv[i], "--noborder")) 
				fWorkspacesWindow->SetLook(B_NO_BORDER_WINDOW_LOOK);
			else if (! strcmp(argv[i], "--acceptfirstclick"))
				fWorkspacesWindow->SetFlags(B_WILL_ACCEPT_FIRST_CLICK);
			else if (! strcmp(argv[i], "--avoidfocus"))
				fWorkspacesWindow->SetFlags(B_AVOID_FOCUS);
			else if (! strcmp(argv[i], "--alwaysontop"))
				fWorkspacesWindow->SetFeel(B_FLOATING_ALL_WINDOW_FEEL);
			else if (! strcmp(argv[i], "--version")) 
				Version();
			else if (! strcmp(argv[i], "--help"))
				Usage();
			else /* if ( argv[i][0] == '-') */
				SuggestHelpMessage();
		}  else if ( isdigit(*argv[i]) ) {
			// check for a numeric arg, if not already given
			activate_workspace(atoi(argv[i]));
			// if the app is running, don't quit
			// but if it isn't, cancel the complete run, so it doesn't
			// open any window
			if (IsLaunching())
				Quit();
		} else {
			//else use that as app then args to that app
			BMessage msg(B_OPEN_IN_WORKSPACE);
			BEntry e(argv[i]); 
			entry_ref ref;
			if ((e.InitCheck() == B_OK) && (e.GetRef(&ref) == B_OK)) {
				if ( isdigit(*argv[i-1]) )
					msg.AddInt32("be:workspace", atoi(argv[i-1])); 
				be_roster->Launch(&ref, &msg);
			} 
		}
	}

}

void
WorkspacesApp::ReadyToRun(void)
{
	(fWorkspacesWindow)->Show();
}


bool
WorkspacesApp::QuitRequested(void)
{
	return true;
}


//	#pragma mark -


int
main(int argc, char **argv)
{
	new WorkspacesApp();
	be_app->Run();
	delete be_app;
	return 0;
}

